﻿#pragma TextEncoding="UTF-8"
#pragma rtGlobals=3
#pragma IgorVersion=8
#pragma version=1.10
#pragma ModuleName=MarqueePeaks

// --------------------- Project Updater header ----------------------
// If you have installed or updated this project using the IgorExchange 
// Projects Installer (http: www.igorexchange.com/project/Updater) you 
// can be notified when new versions are released.
static constant kProjectID=21526 // the project node on IgorExchange
static strconstant ksShortTitle="Marquee Peaks" // the project short title on IgorExchange

// Procedures to help make some common tasks easy.
// Most of these use the graph marquee to select a subrange of a plotted trace.
// Used to determine peak heights, areas, centroids (centres of mass), and FWHM.
// Right click within the graph marquee to select which of these you want
// to calculate. When a trace is selected from the list of traces that
// pass though the marquee, the results of the calculation are written to
// history and shown in a tag on the graph. Select the 'clear' submenu to
// remove tags from the plot.

// WAVES MUST BE BASELINE-FREE FOR THESE FUNCTIONS TO BE USEFUL
// fit-derived uncertainties are not really meaningful.

// Peak height: fits a third order polynomial through the selected range,
// and then tries to find the peak by looking for the first zero-gradient
// within that selected range. For this to work nicely you should select a
// narrow region around the peak. I used a polynomial so that it should
// work with asymmetric peaks. Also gives peak position.

// Area: simply calculates the area within the selected x-range.

// Centroid: integrates the data over the selected range and looks
// for the position of area/2 in the integral wave. Also gives area.

// Peak fit: First fits a Gaussian and a Lorentzian over the selected 
// range, then uses the fit coefficients as starting guesses to fit a 
// Voigt profile over the same range. Gives peak height, FWHM and 
// position for the best fit.

// Attempts to fit a pair of peaks to a doublet and derives peak 
// position, height and FWHM parameters. Works best if the peaks are 
// positioned at 1/4 and 3/4 of the selected x-range. Selects the best 
// fit from attempts to fit a pair of Lorentzian, Gaussian and Voigt 
// peaks.

// Includes options to do all-in-one determinations for all of the
// traces on the plot. The results of all-in-one calculation are
// presented in a table. The waves displayed in the table are located in
// the package folder. It's up to the user to copy data into a more
// useful location.

// This file also has some functions for normalizing spectra by peak 
// height, peak area or total area. To normalize by total area or 
// wavemax, use the trace and all traces menus (right click or 
// shift-right click on a trace).

// 'Print waves in marquee' prints the names of marquee traces in the
// history

// https://www.wavemetrics.com/user/tony

// edit here to exclude traces from the marquee menus
strconstant ksExclude = "fit_*;MQP_*;"

menu "AllTracesPopup"
	"Normalize All Traces By Total Area", /Q, MarqueePeaks#NormaliseAllTracesByArea()
	"Normalize All Traces By Maximum Value", /Q, MarqueePeaks#NormaliseAllTracesByMax()
end

menu "TracePopup"
	"Normalize By Total Area", /Q, MarqueePeaks#NormaliseTraceByArea()
	"Normalize By Wave Maximum", /Q, MarqueePeaks#NormaliseTraceByMax()
end

// -----------------  marquee menu functions for peak height and area ------------------

menu "GraphMarquee", dynamic
	submenu "Peak Height (poly fit)"
		MarqueePeaks#TracesInMarquee(exclude=ksExclude), MarqueePeaks#MarqueeHeight()
		"-"
		"All Traces", MarqueePeaks#MarqueeHeight()
		"Clear", MarqueePeaks#ResetGraph("")
	end
	submenu "Marquee Area"
		MarqueePeaks#TracesInMarquee(exclude=ksExclude), MarqueePeaks#MarqueeArea()
		"-"
		"All Traces", MarqueePeaks#MarqueeArea()
		"Clear", MarqueePeaks#ResetGraph("")
	end
	submenu "Marquee Centroid"
		MarqueePeaks#TracesInMarquee(exclude=ksExclude), MarqueePeaks#MarqueeCentroid()
		"-"
		"All Traces", MarqueePeaks#MarqueeCentroid()
		"Clear", MarqueePeaks#ResetGraph("")
	end
	submenu "Fit Peak"
		MarqueePeaks#TracesInMarquee(exclude=ksExclude), MarqueePeaks#MarqueePeak()
		"-"
		"All Traces", MarqueePeaks#MarqueePeak()
		"Clear", MarqueePeaks#ResetGraph("")
	end
	submenu "Fit Doublet"
		MarqueePeaks#TracesInMarquee(exclude=ksExclude), MarqueePeaks#MarqueeDoublet()
		"-"
		"All Traces", MarqueePeaks#MarqueeDoublet()
		"Clear", MarqueePeaks#ResetGraph("")
	end
	submenu "Normalise By Selected Area"
		MarqueePeaks#TracesInMarquee(exclude=ksExclude),/Q, MarqueePeaks#MarqueeNormaliseTraces()
		"-"
		"All Traces", MarqueePeaks#MarqueeNormaliseTraces()
	end
	submenu "Normalise By Marquee Max"
		MarqueePeaks#TracesInMarquee(exclude=ksExclude), MarqueePeaks#MarqueeNormaliseByMax()
		"-"
		"All Traces", MarqueePeaks#MarqueeNormaliseByMax()
	end
	"Print Waves In Marquee", Print MarqueePeaks#TracesInMarquee()
end

static function /S TracesInMarquee([string exclude])
	if (ParamIsDefault(exclude))
		exclude = ""
	else
		exclude += ";"
		exclude = ReplaceString(";;", exclude, ";")
	endif
	GetMarquee /Z
	string inlist = TraceNameList(S_marqueeWin,";",5) // excludes hidden traces
	string outlist = ""
	string s_trace, yAxisName, xAxisName, s_info
	variable xOffset, yOffset, xMult, yMult
	int i, j, ignore
	int ntraces = ItemsInList(inlist)
	for(i=0;i<ntraces;i+=1)
		s_trace = StringFromList(i,inlist)
		// check for waves to ignore
		ignore = 0
		for (j=0;j<ItemsInList(exclude);j+=1)
			if (stringmatch(s_trace, StringFromList(j, exclude)))
				ignore = 1
				break
			endif
		endfor
		if (ignore)
			continue
		endif
		wave /Z w = TraceNameToWaveRef(S_marqueeWin, s_trace)
		s_info = TraceInfo(S_marqueeWin,s_trace,0)
		xAxisName = StringByKey("XAXIS", s_info)
		yAxisName = StringByKey("YAXIS", s_info)	
		GetMarquee /Z $xAxisName, $yAxisName
		
		// deal with reversed axes by making v_top > v_bottom
		// and v_right > v_left
		variable swap = max(v_bottom, v_top)
		v_bottom = min(v_bottom, v_top)
		v_top = swap
		swap = max(v_left, v_right)
		v_left = min(v_left, v_right)
		v_right = swap
		
		// get trace offsets
		xOffset = GetNumFromModifyStr(s_info,"offset","{",0)
		yOffset = GetNumFromModifyStr(s_info,"offset","{",1)
		
		v_left -= xOffset
		v_right -= xOffset
		v_top -= yOffset
		v_bottom -= yOffset
		
		xMult = GetNumFromModifyStr(s_info,"muloffset","{",0)
		yMult = GetNumFromModifyStr(s_info,"muloffset","{",1)
		
		v_right = (xMult!=0) ? v_right/xMult : v_right
		v_left = (xMult!=0) ? v_left/xMult : v_left
		v_top = (yMult!=0) ? v_top/yMult : v_top
		v_bottom = (yMult!=0) ? v_bottom/yMult : v_bottom

// ignore mutiplicative offsets.	
//		if (xMult)
//			v_right/=xMult
//			v_left/=xMult
//		endif
//		if (yMult)
//			v_top/=yMult
//			v_bottom/=yMult
//		endif
		
		// deal with x-y data
		wave /z w_x = XWaveRefFromTrace(S_marqueeWin, s_trace)
		if (WaveExists(w_x))
			// be careful not to choke on category plot traces etc.
			if ((WaveType(w_x)==0) || (numpnts(w_x)==0))
				continue
			endif
			int errL = 0, errR = 0
			FindLevel /P/Q w_x, v_left
			if (V_flag)
				errL = 1
				FindLevel /P/Q w_x, WaveMin(w_x)
			endif
			v_left = pnt2x(w,V_LevelX)
			// v_left is the wave x value for leftmost (lowest x) point of w in marquee x-range
			// takes into account scaling of w		
			FindLevel /P/Q w_x, v_right
			if (V_flag)
				errR = 1
				FindLevel /P/Q w_x, WaveMax(w_x)
			endif
			v_right = pnt2x(w,V_LevelX)
			if (errL && errR)
				// wave is plotted to left or right of marquee
				continue
			endif
			// make sure w has unit scaling so that the point numbers (v_left
			// and v_right) correspond to x values for w
		endif
			
		// check for waveform waves plotted left or right of marquee
		if (WaveExists(w_x)==0)
			variable x_low, x_high
			x_low = min(leftx(w), pnt2x(w,numpnts(w)-1))
			x_high = max(leftx(w), pnt2x(w,numpnts(w)-1))
			if (x_low>v_right || x_high<v_left)
				continue
			endif
		endif
				
		// set y_max and y_min to max and min of points plotted
		// within the x-axis range of marquee
		variable y_max = WaveMax(w, v_left, v_right)
		variable y_min = WaveMin(w, v_left, v_right)
				
		if (y_min>v_top || y_max<v_bottom)
			// wave is plotted above or below marquee
			continue
		endif
		
		outlist = AddListItem(StringFromList(i,inlist), outlist)
	endfor
	return outlist
end

static function ResetGraph(string strWin)
	string tracelist = TraceNameList(strWin, ";", 1)
	string trace
	int i
	for(i=ItemsInList(tracelist)-1;i>=0;i--)
		trace = StringFromList(i, tracelist)
		wave w = TraceNameToWaveRef(strWin, trace)
		if (DataFolderRefsEqual(GetWavesDataFolderDFR(w), packageFolder()))
		//if (stringmatch(trace, "MQP_*"))
			RemoveFromGraph /W=$strWin/Z $trace
			KillWaves /z w
		endif
	endfor
	Tag /W=$strWin/K/N=MarqueePeakTag
	Tag /W=$strWin/K/N=MarqueePeakTag2
end

static function /DF packageFolder()
	NewDataFolder /O root:Packages
	NewDataFolder /O root:Packages:MarqueePeaks
	return root:Packages:MarqueePeaks
end

static function MarqueeHeight()
	GetMarquee /Z
	if (v_flag==0)
		return 0
	endif
	string strGraph = S_marqueeWin
	ResetGraph(S_marqueeWin)
	string tracename = "", tracelist = "", strIgnore = "", strXAxis = ""
	GetLastUserMenuInfo // S_value will be trace name
	tracelist = SelectString(cmpstr(S_value, "All Traces")==0, s_value, TraceNameList("",";",1))
	int multiTrace = ItemsInList(tracelist)>1
	if (multiTrace)
		DFREF dfr = packageFolder()
		wave /SDFR=dfr/Z w_peakheight, w_position
		wave /SDFR=dfr/Z/T w_name
		if (WaveExists(w_peakheight)||WaveExists(w_name)||WaveExists(w_position))
			DoAlert 1, "w_name/w_peakheight/w_position exists. Overwrite?"
			if (V_flag == 2)
				Print "multi-peak determination aborted"
				return 0
			endif
		endif
		Make /O/N=0 dfr:w_peakheight /Wave=w_peakheight, dfr:w_position /Wave=w_position
		Make /O/T/N=0 dfr:w_name /Wave=w_name
	endif
	Make /free/N=2 xMinMax
	
	int numTraces = ItemsInList(tracelist), i = 0
	for(i=0;i<numTraces;i+=1)
		traceName = StringFromList(i,tracelist)
		if (stringmatch(traceName, "fit_*") || stringmatch(traceName, "MQP_*")) // ignore fit waves
			strIgnore += tracename + ", "
		else
			wave w = TraceNameToWaveRef("",tracename)
			wave /Z w_x = XWaveRefFromTrace("",tracename)
			strXAxis = StringByKey("XAXIS", TraceInfo("", tracename, 0 ))
			GetMarquee $strXAxis
			xMinMax = {v_left,v_right}
			if (withinRange(w, w_x, xMinMax)==0)
				continue
			endif
			wave wOut = peakHeight(w, w_x, xMinMax, strGraph, tracename, multiTrace==0)
			if (multiTrace==0) // done
				GetMarquee /Z/K
				return 1
			endif					
			if (numtype(wOut[0])==0)
				InsertPoints numpnts(w_name), 1, w_peakheight, w_position, w_name
				w_peakheight[numpnts(w_peakheight)-1] = wOut[0]
				w_position[numpnts(w_position)-1] = wOut[1]
				w_name[numpnts(w_name)-1] = NameOfWave(w)
				Make /O/N=300 dfr:$"MQP_Fit_"+NameOfWave(w) /Wave=MQP_Fit
				SetScale /I x, xMinMax[0], xMinMax[1], MQP_Fit
				DeletePoints 0, 2, wOut
				MQP_Fit = poly(wOut,x)
				appendToSameAxes(strGraph, tracename, MQP_Fit, $"", offset=1, lsize=2)
			endif
		endif
	endfor
	
	if (strlen(strIgnore))
		strIgnore = RemoveEnding(strIgnore  , ", ")
		printf "Ignoring traces %s\r", strIgnore
	endif
	
	if (numpnts(w_peakheight))
		KillWindow /Z MarqueePeakTable
		Edit /N=MarqueePeakTable/K=1 w_name, w_position, w_peakheight
	endif
	GetMarquee /Z/K
	return 1
end

static function MarqueeArea()
	GetMarquee /Z
	if (v_flag==0)
		return 0
	endif
	string strGraph = S_marqueeWin
	ResetGraph(S_marqueeWin)
	
	string tracename = "", tracelist = "", strIgnore = "", strXAxis = ""
	
	GetLastUserMenuInfo // S_value will be trace name
	tracelist = SelectString(cmpstr(S_value, "All Traces")==0, s_value, TraceNameList("",";",1))
	int multiTrace = ItemsInList(tracelist)>1
	if (multiTrace)
		DFREF dfr = packageFolder()
		wave /SDFR=dfr/Z w_area
		wave /SDFR=dfr/Z/T w_name
		if (WaveExists(w_area)||WaveExists(w_name))
			DoAlert 1, "w_name/w_area exists. Overwrite?"
			if (V_flag == 2)
				Print "multi-area determination aborted"
				return 0
			endif
		endif
		Make /O/N=0 dfr:w_area /Wave=w_area
		Make /O/T/N=0 dfr:w_name /Wave=w_name
	endif
	Make /free/N=2 xMinMax
	
	variable vArea
	int numTraces = ItemsInList(tracelist)
	int i
	for(i=0;i<numTraces;i+=1)
		traceName = StringFromList(i,tracelist)
		if (stringmatch(traceName, "fit_*") || stringmatch(traceName, "MQP_*")) // ignore fit waves
			strIgnore += tracename + ", "
		else
			wave w = TraceNameToWaveRef("",tracename)
			wave /Z w_x = XWaveRefFromTrace("",tracename)
			strXAxis = StringByKey("XAXIS", TraceInfo("", tracename, 0 ))
			GetMarquee $strXAxis
			xMinMax = {v_left,v_right}
			if (withinRange(w, w_x, xMinMax)==0)
				continue
			endif
			
			vArea = peakArea(w, w_x, xMinMax, strGraph, tracename, multiTrace==0)
			if (multiTrace==0) // done
				GetMarquee /Z/K
				return 1
			endif
					
			if (numtype(vArea)==0)
				InsertPoints numpnts(w_name), 1, w_area, w_name
				w_area[numpnts(w_area)-1] = vArea
				w_name[numpnts(w_name)-1] = NameOfWave(w)
			endif
			
			Make /O/N=2 dfr:MQP_Area /wave=MQP_Area
			SetScale /I x, xMinMax[0], xMinMax[1], MQP_Area
			MQP_Area = Inf
			appendToSameAxes(strGraph, tracename, MQP_Area, $"", offset=1, fill=1)			
		endif
	endfor
	
	if (strlen(strIgnore))
		strIgnore = RemoveEnding(strIgnore  , ", ")
		printf "Ignoring traces %s\r", strIgnore
	endif
	
	if (numpnts(w_name))
		KillWindow /Z MarqueePeakTable
		Edit /N=MarqueePeakTable/K=1 w_name, w_area
	endif
	GetMarquee /Z/K
	return 1
end

static function MarqueeCentroid()
	GetMarquee /Z
	if (v_flag==0)
		return 0
	endif
	string strGraph = S_marqueeWin
	ResetGraph(S_marqueeWin)
	
	string tracename = "", tracelist = "", strIgnore = "", strXAxis = ""
	
	GetLastUserMenuInfo // S_value will be trace name
	tracelist = SelectString(cmpstr(S_value, "All Traces")==0, s_value, TraceNameList("",";",1))
	int multiTrace = ItemsInList(tracelist)>1
	if (multiTrace)
		DFREF dfr = packageFolder()
		wave /SDFR=dfr/Z w_centroid
		wave /SDFR=dfr/Z/T w_name
		if (WaveExists(w_centroid)||WaveExists(w_name))
			DoAlert 1, "w_name/w_centroid exists. Overwrite?"
			if (V_flag == 2)
				Print "multi-centroid determination aborted"
				return 0
			endif
		endif
		Make /O/N=0 dfr:w_centroid /Wave=w_centroid
		Make /O/T/N=0 dfr:w_name /Wave=w_name
	endif
	Make /free/N=2 xMinMax
	
	variable centroid
	int numTraces = ItemsInList(tracelist)
	int i
	for(i=0;i<numTraces;i+=1)
		traceName=StringFromList(i,tracelist)
		if (stringmatch(traceName, "fit_*") || stringmatch(traceName, "MQP_*")) // ignore fit waves
			strIgnore += tracename + ", "
		else
			wave w = TraceNameToWaveRef("",tracename)
			wave /Z w_x = XWaveRefFromTrace("",tracename)
			strXAxis = StringByKey("XAXIS", TraceInfo("", tracename, 0 ))
			GetMarquee $strXAxis
			xMinMax = {v_left,v_right}
			if (withinRange(w, w_x, xMinMax)==0)
				continue
			endif
			
			centroid = peakCentroid(w, w_x, xMinMax, strGraph, tracename, multiTrace==0)
			if (multiTrace==0) // done
				GetMarquee /Z/K
				return 1
			endif
					
			if (numtype(centroid)==0)
				InsertPoints numpnts(w_name), 1, w_centroid, w_name
				w_centroid[numpnts(w_centroid)-1] = centroid
				w_name[numpnts(w_name)-1] = NameOfWave(w)
			endif
			
			Make /O/N=2 dfr:MQP_Area /wave=MQP_Area
			SetScale /I x, xMinMax[0], xMinMax[1], MQP_Area
			MQP_Area = Inf
			appendToSameAxes(strGraph, tracename, MQP_Area, $"", offset=1, fill=1)
		endif
	endfor
	
	if (strlen(strIgnore))
		strIgnore = RemoveEnding(strIgnore  , ", ")
		printf "Ignoring traces %s\r", strIgnore
	endif
	
	if (numpnts(w_name))
		KillWindow /Z MarqueePeakTable
		Edit /N=MarqueePeakTable/K=1 w_name, w_centroid
	endif
	GetMarquee /Z/K
	return 1
end

static function MarqueeDoublet()
	GetMarquee /Z
	if (v_flag==0)
		return 0
	endif
	string strGraph = S_marqueeWin
	ResetGraph(S_marqueeWin)
	
	string tracename = "", tracelist = "", strIgnore = "", strXAxis = ""
	
	GetLastUserMenuInfo // S_value will be trace name
	tracelist = SelectString(cmpstr(S_value, "All Traces")==0, s_value, TraceNameList("",";",1))
	int multiTrace = ItemsInList(tracelist)>1
	if (multiTrace)
		DFREF dfr = packageFolder()
		wave /SDFR=dfr/Z w_peakheight, w_position, w_FWHM
		wave /SDFR=dfr/Z/T w_name
		if (WaveExists(w_peakheight)||WaveExists(w_name)||WaveExists(w_position)||WaveExists(w_FWHM))
			DoAlert 1, "w_name/w_peakheight/w_position/w_FWHM exists. Overwrite?"
			if (V_flag == 2)
				Print "multi-trace doublet fitting aborted"
				return 0
			endif
		endif
		Make /O/N=0 dfr:w_peakheight /Wave=w_peakheight, dfr:w_position /Wave=w_position, dfr:w_FWHM /Wave=w_FWHM
		Make /O/T/N=0 dfr:w_name /Wave=w_name
	endif
	Make /free/N=2 xMinMax
	
	int numTraces = ItemsInList(tracelist), i = 0
	for(i=0;i<numTraces;i+=1)
		traceName = StringFromList(i,tracelist)
		if (stringmatch(traceName, "fit_*") || stringmatch(traceName, "MQP_*")) // ignore fit waves
			strIgnore += tracename + ", "
		else
			wave w = TraceNameToWaveRef("",tracename)
			wave /Z w_x = XWaveRefFromTrace("",tracename)
			strXAxis = StringByKey("XAXIS", TraceInfo("", tracename, 0 ))
			GetMarquee $strXAxis
			xMinMax = {v_left,v_right}
			if (withinRange(w, w_x, xMinMax)==0)
				continue
			endif
			wave wOut = PeakFitDoublet(w, w_x, xMinMax, strGraph, tracename, multiTrace==0)
			if (multiTrace==0) // done
				GetMarquee /Z/K
				return 1
			endif
			
			if (numpnts(wOut)==6)
				InsertPoints numpnts(w_name), 2, w_peakheight, w_position, w_name, w_FWHM
				w_peakheight[numpnts(w_peakheight)-2] = wOut[0]
				w_position[numpnts(w_position)-2] = wOut[1]
				w_FWHM[numpnts(w_FWHM)-2] = wOut[2]
				w_name[numpnts(w_name)-2] = NameOfWave(w)
				
				w_peakheight[numpnts(w_peakheight)-1] = wOut[3]
				w_position[numpnts(w_position)-1] = wOut[4]
				w_FWHM[numpnts(w_FWHM)-1] = wOut[5]
				w_name[numpnts(w_name)-1] = NameOfWave(w)
				
				wave MQP_Fit = dfr:MQP_Fit
				Duplicate /O MQP_Fit dfr:$"MQP_Fit" + NameOfWave(w) /wave=MQP_Fit
				appendToSameAxes(strGraph, tracename, MQP_Fit, $"", offset=1, lsize=2)
			endif
		endif
	endfor
	
	if (strlen(strIgnore))
		strIgnore = RemoveEnding(strIgnore  , ", ")
		printf "Ignoring traces %s\r", strIgnore
	endif
	
	if (numpnts(w_name))
		KillWindow /Z MarqueePeakTable
		Edit /N=MarqueePeakTable/K=1 w_name, w_position, w_peakheight, w_FWHM
	endif
	GetMarquee /Z/K
	return 1
end

static function MarqueePeak()
	GetMarquee /Z
	if (v_flag==0)
		return 0
	endif
	string strGraph = S_marqueeWin
	ResetGraph(S_marqueeWin)
	
	string tracename = "", tracelist = "", strIgnore = "", strXAxis = ""
	
	GetLastUserMenuInfo // S_value will be trace name
	tracelist = SelectString(cmpstr(S_value, "All Traces")==0, s_value, TraceNameList("",";",1))
	int multiTrace = ItemsInList(tracelist)>1
	if (multiTrace)
		DFREF dfr = packageFolder()
		wave /SDFR=dfr/Z w_peakheight, w_position, w_FWHM
		wave /SDFR=dfr/Z/T w_name
		if (WaveExists(w_peakheight)||WaveExists(w_name)||WaveExists(w_position)||WaveExists(w_FWHM))
			DoAlert 1, "w_name/w_peakheight/w_position/w_FWHM exists. Overwrite?"
			if (V_flag == 2)
				Print "multi-trace peak fitting aborted"
				return 0
			endif
		endif
		Make /O/N=0 dfr:w_peakheight /Wave=w_peakheight, dfr:w_position /Wave=w_position, dfr:w_FWHM /Wave=w_FWHM
		Make /O/T/N=0 dfr:w_name /Wave=w_name
	endif
	Make /free/N=2 xMinMax
	
	int numTraces = ItemsInList(tracelist), i = 0
	for(i=0;i<numTraces;i+=1)
		traceName = StringFromList(i,tracelist)
		if (stringmatch(traceName, "fit_*") || stringmatch(traceName, "MQP_*")) // ignore fit waves
			strIgnore += tracename + ", "
		else
			wave w = TraceNameToWaveRef("",tracename)
			wave /Z w_x = XWaveRefFromTrace("",tracename)
			strXAxis = StringByKey("XAXIS", TraceInfo("", tracename, 0 ))
			GetMarquee $strXAxis
			xMinMax = {v_left,v_right}
			if (withinRange(w, w_x, xMinMax)==0)
				continue
			endif
			wave wOut = PeakFit(w, w_x, xMinMax, strGraph, tracename, multiTrace==0)
			if (multiTrace==0) // done
				GetMarquee /Z/K
				return 1
			endif
			
			if (numpnts(wOut)==3)
				InsertPoints numpnts(w_name), 1, w_peakheight, w_position, w_name, w_FWHM
				w_peakheight[numpnts(w_peakheight)-1] = wOut[0]
				w_position[numpnts(w_position)-1] = wOut[1]
				w_FWHM[numpnts(w_FWHM)-1] = wOut[2]
				w_name[numpnts(w_name)-1] = NameOfWave(w)
				wave MQP_Fit = dfr:MQP_Fit
				Duplicate /O MQP_Fit dfr:$"MQP_Fit" + NameOfWave(w) /wave=MQP_Fit
				appendToSameAxes(strGraph, tracename, MQP_Fit, $"", offset=1, lsize=2)
			endif
		endif
	endfor
	
	if (strlen(strIgnore))
		strIgnore = RemoveEnding(strIgnore  , ", ")
		printf "Ignoring traces %s\r", strIgnore
	endif
	
	if (numpnts(w_name))
		KillWindow /Z MarqueePeakTable
		Edit /N=MarqueePeakTable/K=1 w_name, w_position, w_peakheight, w_FWHM
	endif
	GetMarquee /Z/K
	return 1
end

// truth that at least part of w [vs w_x] is within x range specified by 2 point wave 'range'
static function withinRange(wave w, wave /Z w_x, wave range)
	variable xmin, xmax, wmin, wmax
	wmin = WaveExists(w_x) ? WaveMin(w_x) : min(pnt2x(w, 0), pnt2x(w, numpnts(w)-1))
	wmax = WaveExists(w_x) ? WaveMax(w_x) : max(pnt2x(w, 0), pnt2x(w, numpnts(w)-1))
	if (wmin>WaveMax(range) || wmax<WaveMin(range))
		return 0
	endif
	return 1
end

static function /wave peakHeight(wave w, wave /Z w_x, wave xMinMax, string strGraph, string tracename, int doTag)
	variable xmin = WaveMin(xMinMax), xmax = WaveMax(xMinMax)
	wave w_EndPoints = getEndPoints(w, w_x, xMinMax)
	Make /d/free/O/N=4 cw // no need for good initial guesses for poly coefficients
	Make /d/free wOut = {NaN, NaN}

	int sav_debug = DebuggerOff()
	try
		CurveFit /Q/NTHR=0/TBOX=0 poly 4, kwCWave=cw, w[w_EndPoints[0],w_EndPoints[1]] /X=w_x/NWOK; AbortOnRTE
	catch
		variable error = GetRTError(1)
	endtry
	DebuggerOptions debugOnError = sav_debug
	if (error)
		return wOut
	endif

	if (doTag)
		DFREF dfr = packageFolder()
		Make /O/N=300 dfr:MQP_Fit /Wave=MQP_Fit
		SetScale /I x, xMinMax[0], xMinMax[1], MQP_Fit
		MQP_Fit = poly(cw,x)
		appendToSameAxes(strGraph, tracename, MQP_Fit, $"", offset=1)
	endif
	// differentiate the poly
	Make /D/free/N=3 wPoly3 = {cw[1],2*cw[2],3*cw[3]}
	FindRoots /P=wPoly3
	wave W_polyRoots = W_polyRoots
	variable x0 = NaN, y0 = -Inf
	int i
	for(i=0;i<numpnts(W_polyRoots);i+=1)
		if (imag(W_polyRoots[i])!=0)
			continue
		endif
		if (real(W_polyRoots[i])>xmin && real(W_polyRoots[i])<xmax)
			// found a root
			if (poly(cw, W_polyRoots[i])>y0)
				x0 = W_polyRoots[i]
				y0 = poly(cw, x0)
			endif
		endif
	endfor
	if (doTag)
		if (numtype(x0)==0)
			string strTag = ""
			sprintf strTag "%s\rHeight = %g\rPos = %g", NameOfWave(w), poly(cw, x0), x0
			Tag /A=MB/B=1/C/N=MarqueePeakTag/I=0/L=2/Z=0/X=0/Y=5 MQP_Fit, x0, strTag
		else
			Tag /K/N=MarqueePeakTag
		endif
	endif
	printf "%s: range = %g-%g, x0 = %g, y0 = %g\r", NameOfWave(w), xMinMax[0], xMinMax[1], x0, poly(cw, x0)
	wOut = {y0, x0, cw[0], cw[1], cw[2], cw[3]}
	return wOut
end

static function peakArea(wave w, wave /Z w_x, wave xMinMax, string strGraph, string tracename, int doTag)
	DFREF dfr = packageFolder()
	wave /Z MQP_AreaX = $""
	if (WaveExists(w_x))
		wave w_EndPoints = getEndPoints(w, w_x, xMinMax)
		Duplicate /O /R=[w_EndPoints[0], w_EndPoints[1]] w dfr:MQP_Area /wave=MQP_Area
		Duplicate /O /R=[w_EndPoints[0], w_EndPoints[1]] w_x dfr:MQP_AreaX /wave=MQP_AreaX
	else
		Duplicate /O /R=(xMinMax[0], xMinMax[1]) w dfr:MQP_Area /wave=MQP_Area
	endif
	variable vArea = AreaXoptional(MQP_Area,MQP_AreaX,-Inf,Inf)
	if (doTag)
		appendToSameAxes(strGraph, tracename, MQP_Area, MQP_AreaX, offset=1, lsize=0, fill=1)
		string strTag
		sprintf strTag "%s\rArea = %g", NameOfWave(w), vArea
		WaveStats /M=1/Q MQP_Area
		Tag /A=MB/B=1/C/N=MarqueePeakTag/I=0/L=2/Z=0/X=0/Y=5 $tracename, V_maxloc, strTag
	endif
	printf "%s: range %g-%d, area = %g\r", NameOfWave(w), xMinMax[0], xMinMax[1], vArea
	return vArea
end
	
static function peakCentroid(wave w, wave /Z w_x, wave xMinMax, string strGraph, string tracename, int doTag)
	DFREF dfr = packageFolder()
	// edit here to expand to zero crossing points
	int ExpandToZeroCrossing = 0
	
	variable vLeft = xMinMax[0], vRight = xMinMax[1]
	if (WaveExists(w_x))
		// figure out marquee horizontal dimensions in terms of x scaling for w
		if (DimDelta(w_x,0)!=DimDelta(w,0) || DimOffset(w_x,0)!=DimOffset(w,0))
			printf "%s: centroid calculation requires same scaling for X and Y waves\r", NameOfWave(w)
			return NaN
		endif
		FindLevel /Q w_x, vLeft
		vLeft = V_LevelX
		FindLevel /Q w_x, vRight
		vRight = V_LevelX
	endif
	
	if (ExpandToZeroCrossing)
		variable inc = DimDelta(w,0)/4
		if (w(vRight)<0 || w(vLeft)<0)
			printf "%s: left and right bounds must be above zero!\r", NameOfWave(w)
			if(doTag)
				DoAlert 0, "left and right bounds must be above zero!"
			endif
			return NaN
		endif
		do
			vRight += inc
		while (w(vRight)>0 && vRight<max(rightx(w), leftx(w)) && vRight>min(rightx(w), leftx(w)))
		do
			vLeft -= inc
		while (w(vLeft)>0 && vLeft<max(rightx(w), leftx(w)) && vLeft>min(rightx(w), leftx(w)))
	endif
	
	wave /Z MQP_AreaX = $""
	Duplicate /O /R=(vLeft, vRight) w dfr:MQP_Area /wave=MQP_Area
	if (WaveExists(w_x))
		Duplicate /O /R=(vLeft, vRight) w_x dfr:MQP_AreaX /wave=MQP_AreaX
	endif	
	Make /free PeakIntWave
	if (WaveExists(w_x))
		Integrate /METH=1 MQP_Area /D=PeakIntWave /X=MQP_AreaX
	else
		Integrate /METH=1 MQP_Area /D=PeakIntWave
	endif
	PeakIntWave *= sign(PeakIntWave[Inf])	
	variable vArea = PeakIntWave[Inf]
	FindLevel /Q PeakIntWave, vArea/2
	if (V_flag)
		return NaN
	endif
	variable centroid = WaveExists(MQP_AreaX) ? MQP_AreaX(V_LevelX) : V_LevelX
	if (doTag)
		string strTag
		appendToSameAxes(strGraph, tracename, MQP_Area, MQP_AreaX, offset=1, lsize=0, fill=1)
		sprintf strTag "%s\rCentroid = %g\rTotal area = %g", NameOfWave(w), centroid, vArea
		Tag /A=MB/B=1/C/N=MarqueePeakTag/I=0/L=2/Z=0/X=0/Y=5 $tracename, V_LevelX, strTag
	endif
	printf "%s: range %g-%d, centroid = %g, total area = %g\r", NameOfWave(w), xMinMax[0], xMinMax[1], centroid, vArea
	return centroid
end

static function /wave PeakFit(wave w, wave/Z w_x, wave xMinMax, string strGraph, string tracename, int doTag)
	DFREF dfr = packageFolder()
	wave w_EndPoints = getEndPoints(w, w_x, xMinMax)
	int sav_debug = DebuggerOff()
	string strTag
	variable gChiSq, lChiSq, vChiSq, gError, lError, vError, wg, wl, y0
	variable xmin = WaveMin(xMinMax), xmax = WaveMax(xMinMax)
	
	WaveStats /Q/R=[w_EndPoints[0],w_EndPoints[1]] w
	Make /O/D/free cwg = {v_max, xmin+(xmax-xmin)/2, (xmax-xmin)/2}
	// set constraints for Gaussian and Lorentzian fit
	Make /O/free/T T_Constraints = {"K0>0","K0<"+num2str(1.3*V_max),"K1>"+num2str(xmin),"K1<"+num2str(xmax),"K2>0"}
	
	// first fit a Gaussian
	try
		FuncFit /Q/NTHR=0/TBOX=0 MarqueePeaks#gPeak, kwCWave=cwg, w[w_EndPoints[0], w_EndPoints[1]] /X=w_x/NWOK/C=T_Constraints; AbortOnRTE
		wave w_sigma
		Duplicate /O/free w_sigma swg
		gChiSq = v_chiSq
	catch
		gError = GetRTError(1)
	endtry
	
	// now try Lorentzian
	Make /O/D/free cwl = {v_max, xmin+(xmax-xmin)/2, (xmax-xmin)/2}
	try
		FuncFit /Q/NTHR=0/TBOX=0 MarqueePeaks#lPeak, kwCWave=cwl, w[w_EndPoints[0], w_EndPoints[1]] /X=w_x/NWOK/C=T_Constraints; AbortOnRTE
		// now fit a Gaussian with no y offset
		wave w_sigma
		Duplicate /O/free w_sigma swl
		lChiSq = v_chiSq
	catch
		lError = GetRTError(1)
	endtry
	
	// and now Voigt
	try
		Make /D/free/O/N=4 cwv
		if (gError==0)
			cwv = {cwg[0]*cwg[2]*sqrt(Pi/(4*ln(2))), cwg[1], cwg[2], 1}
		else
			cwv = {AreaXoptional(w, w_x, xmin, xmax), xmin+(xmax-xmin)/2, (xmax-xmin)/2, 1}
		endif
		Make /free/T/O T_Constraints = {"K0>0","K1>"+num2str(xmin),"K1<"+num2str(xmax),"K2>0","K3>0"}
		FuncFit /Q/NTHR=0/TBOX=0/M=2 MarqueePeaks#vPeakFWHM, kwCWave=cwv, w[w_EndPoints[0], w_EndPoints[1]] /X=w_x/NWOK/C=T_Constraints; AbortOnRTE
		wave w_sigma
		Duplicate /free w_sigma swv
		vChiSq = v_chiSq
	catch
		vError = GetRTError(1)
	endtry
	DebuggerOptions debugOnError = sav_debug // end of fitting
	
	if (vError && lError && gError)
		Make /D/O/free wOut = {NaN,NaN,NaN}
		return wOut
	endif
	
	Make /O/N=300 dfr:MQP_Fit /Wave=MQP_Fit
	SetScale /I x, xMinMax[0], xMinMax[1], MQP_Fit
	if (doTag)
		appendToSameAxes(strGraph, tracename, MQP_Fit, $"", offset=1, fill=1)
	endif
	
	if (gError==0 && (lError || gChiSq<lChisq) && (vError || gChiSq<vChisq) )
		MQP_Fit = gPeak(cwg,x)
		if (doTag)
			sprintf strTag "%s\rFWHM (Guassian) = %g±%g\rx0 = %g±%g\ry0 = %g±%g", NameOfWave(w), cwg[2], swg[2], cwg[1], swg[1], cwg[0], swg[0]
			Tag /A=MB/B=1/C/N=MarqueePeakTag/I=0/L=2/Z=0/X=0/Y=5 MQP_Fit, cwg[1], strTag
			printf "%s: Gaussian peak fit between %g and %g\r", NameOfWave(w), xmin, xmax
		else
			printf "%s: ", NameOfWave(w)
		endif
		printf "x0 = %g±%g, y0 = %g±%g, FWHM = %g±%g\r", cwg[1], swg[1], cwg[0], swg[0], cwg[2], swg[2]
		Duplicate/O/free cwg wOut
	elseif (lError==0 && (gError || lChiSq<gChisq) && (vError || lChiSq<vChisq) )
		MQP_Fit = lPeak(cwl,x)
		if (doTag)
			sprintf strTag "%s\rFWHM (Lorentzian) = %g±%g\rx0 = %g±%g\ry0 = %g±%g", NameOfWave(w), cwl[2], swl[2], cwl[1], swl[1], cwl[0], swl[0]
			Tag /A=MB/B=1/C/N=MarqueePeakTag/I=0/L=2/Z=0/X=0/Y=5 MQP_Fit, cwl[1], strTag
			printf "%s: Lorentzian peak fit between %g and %g\r", NameOfWave(w), xmin, xmax
		else
			printf "%s: ", NameOfWave(w)
		endif
		printf "x0 = %g±%g, y0 = %g±%g, FWHM = %g±%g\r", cwl[1], swl[1], cwl[0], swl[0], cwl[2], swl[2]
		Duplicate/O/free cwl wOut
	elseif (vError==0) // vError should be zero if we arrived here!
		wg = cwv[2] / (0.5346*cwv[3] + sqrt(0.2166*cwv[3]^2 + 1))
		wl = wg * cwv[3]
		y0 = vPeakFWHM(cwv,cwv[1])
		MQP_Fit = vPeakFWHM(cwv,x)
		if (doTag)
			sprintf strTag "%s\rFWHM (Voigt) = %g±%g\rx0 = %g±%g\ry0 = %g", NameOfWave(w), cwv[2], swv[2], cwv[1], swv[1], y0
			Tag /A=MB/B=1/C/N=MarqueePeakTag/I=0/L=2/Z=0/X=0/Y=5 MQP_Fit, cwv[1], strTag
			printf "%s: Voigt peak fit between %g and %g:\r", NameOfWave(w), xmin, xmax
		else
			printf "%s: ", NameOfWave(w)
		endif
		printf "x0 = %g±%g, y0 = %g, FWHM (Voigt) = %g±%g, Gauss FWHM = %g, Lorentzian FWHM = %g, area = %g±%g\r", cwv[1], swv[1], y0, cwv[2], swv[2], wg, wl, cwv[0], swv[0]
		Make /D/O/free wOut={y0, cwv[1], cwv[2]}
	endif
	return wOut
end

static function /wave PeakFitDoublet(wave w, wave/Z w_x, wave xMinMax, string strGraph, string tracename, int doTag)
	DFREF dfr = packageFolder()
	wave w_EndPoints = getEndPoints(w, w_x, xMinMax)
	int sav_debug = DebuggerOff() // switch this off in case the fit fails
	variable xmin = WaveMin(xMinMax), xmax = WaveMax(xMinMax)
	variable gChiSq, lChiSq, vChiSq, gError, lError, vError, wg1, wl1, y01, wg2, wl2, y02
	WaveStats /Q/R=[w_EndPoints[0],w_EndPoints[1]] w
	variable height = v_max/2, FWHM = abs(xMinMax[1]-xMinMax[0])/4
	Make /D/free cwguess = {height, xMinMax[0]+(xMinMax[1]-xMinMax[0])/4, FWHM, height, xMinMax[0]+(xMinMax[1]-xMinMax[0])*3/4, FWHM}
	Duplicate /free cwguess cwg, cwl, cwv
	Make /N=1/free/D wOut = NaN	
	// set constraints for 2 Gaussian and 2 Lorentzian fits
	Make /free/T T_Constraints = {"K0>0","K0<"+num2str(1.3*V_max),"K1>"+num2str(xmin),"K1<"+num2str(xmax),"K2>0"}
	T_Constraints[numpnts(T_Constraints)] = {"K3>0","K3<"+num2str(1.1*V_max),"K4>"+num2str(xmin),"K4<"+num2str(xmax),"K5>0"}	
	// first fit two Gaussians
	try
		FuncFit /Q/NTHR=0/TBOX=0 MarqueePeaks#TwoGaussiansNoBase, kwCWave=cwg, w[w_EndPoints[0], w_EndPoints[1]] /X=w_x/NWOK/C=T_Constraints; AbortOnRTE
	catch
		gError = GetRTError(1)
	endtry
	if (gError==0)
		wave w_sigma = w_sigma
		Duplicate /free w_sigma swg
		gChiSq = V_chisq
	endif
	// now fit two Lorentzians
	try
		FuncFit /Q/NTHR=0/TBOX=0 MarqueePeaks#TwoLorentziansNoBase, kwCWave=cwl, w[w_EndPoints[0], w_EndPoints[1]] /X=w_x/NWOK/C=T_Constraints; AbortOnRTE
	catch
		lError = GetRTError(1)
	endtry
	if (lError==0)
		wave w_sigma = w_sigma
		lChiSq = V_chisq
		Duplicate /free w_sigma swl
	endif
	// use Gaussian fit as starting guesses for Voigt fit
	// for the fit function TwoVoigtNoBase we use parameters area, position, gw, lw for each peak
	// for the fit function TwoVoigtNoBaseFWHM we use parameters area, position, FWHM, l/v for each peak
	// uncertainties for these parameters then come directly from w_sigma	
	if (gError==0)
//		cwv = {cwg[0]*cwg[2]*sqrt(Pi/(4*ln(2))), cwg[1], cwg[2]/2, cwg[2]/2, cwg[3]*cwg[5]*sqrt(Pi/(4*ln(2))), cwg[4], cwg[5]/2, cwg[2]/2}
		cwv = {cwg[0]*cwg[2]*sqrt(Pi/(4*ln(2))), cwg[1], cwg[2], 1, cwg[3]*cwg[5]*sqrt(Pi/(4*ln(2))), cwg[4], cwg[5], 1}
	else // if Gaussian fit wasn't successful, derive guesses from initial guess wave
//		cwv = {cwguess[0]*cwguess[2]*sqrt(Pi/(4*ln(2))), cwguess[1], cwguess[2]/2, cwguess[2]/2, cwguess[3]*cwguess[5]*sqrt(Pi/(4*ln(2))), cwguess[4], cwguess[5]/2, cwguess[2]/2}
		cwv = {cwguess[0]*cwguess[2]*sqrt(Pi/(4*ln(2))), cwguess[1], cwguess[2], 1, cwguess[3]*cwguess[5]*sqrt(Pi/(4*ln(2))), cwguess[4], cwguess[5], 1}
	endif
	T_Constraints = {"K0>0","K1>"+num2str(xmin),"K1<"+num2str(xmax),"K2>0","K3>0"}
	T_Constraints[numpnts(T_Constraints)] = {"K4>0","K5>"+num2str(xmin),"K5<"+num2str(xmax),"K6>0","K7>0"}
	try
		FuncFit /Q/NTHR=0/TBOX=0/M=2 MarqueePeaks#TwoVoigtNoBaseFWHM, kwCWave=cwv, w[w_EndPoints[0], w_EndPoints[1]] /X=w_x/NWOK/C=T_Constraints; AbortOnRTE
	catch
		vError = GetRTError(1)
	endtry
	if (vError==0)
		wave w_sigma = w_sigma
		Duplicate /free w_sigma swv
		vChiSq = V_chisq
	endif
	DebuggerOptions debugOnError = sav_debug // end of fitting
	
	if (gError && lError && vError)
		return wOut
	endif
	
	Make /O/N=300 dfr:MQP_Fit /Wave=MQP_Fit
	Make /O/N=300 dfr:MQP_Fit1 /Wave=MQP_Fit1, dfr:MQP_Fit2 /Wave=MQP_Fit2
	SetScale /i x, xMinMax[0], xMinMax[1], MQP_Fit, MQP_Fit1, MQP_Fit2
	if (doTag)
		appendToSameAxes(strGraph, tracename, MQP_Fit, $"", offset=1, lsize=2)
		appendToSameAxes(strGraph, tracename, MQP_Fit1, $"", offset=1, fill=1)
		appendToSameAxes(strGraph, tracename, MQP_Fit2, $"", offset=1, fill=1)
	endif
	
	Make /D/free cw1, cw2	
	// Uncertainties are really just guesses. No way that these errors are normally distributed.
	// choose the best fit
	if (gError==0 && (gChiSq<lChisq || lError) && (gChiSq<vChisq || vError) )
		MQP_Fit = TwoGaussiansNoBase(cwg,x)
		cw1 = {cwg[0],cwg[1],cwg[2]}
		MQP_Fit1 = gpeak(cw1,x)
		cw2 = {cwg[3],cwg[4],cwg[5]}
		MQP_Fit2 = gpeak(cw2,x)
		printf "%s: Fitting Gaussian peaks between %g and %g\r", NameOfWave(w), xmin, xmax
		printf "Peak 1: x0 = %g±%g, y0 = %g±%g, FWHM = %g±%g\r", cwg[1], swg[1], cwg[0], swg[0], cwg[2], swg[2]
		printf "Peak 2: x0 = %g±%g, y0 = %g±%g, FWHM = %g±%g\r", cwg[4], swg[4], cwg[3], swg[3], cwg[5], swg[5]
	elseif (lError==0 && (lChiSq<gChisq || gError) && (lChiSq<vChisq || vError) )
		MQP_Fit = TwoLorentziansNoBase(cwl,x)
		cw1 = {cwl[0],cwl[1],cwl[2]}
		MQP_Fit1 = lpeak(cw1,x)
		cw2 = {cwl[3],cwl[4],cwl[5]}
		MQP_Fit2 = lpeak(cw2,x)
		printf "%s: Fitting Lorentzian peaks between %g and %g\r", NameOfWave(w), xmin, xmax
		printf "Peak 1: x0 = %g±%g, y0 = %g±%g, FWHM = %g±%g\r", cwl[1], swl[1], cwl[0], swl[0], cwl[2], swl[2]
		printf "Peak 2: x0 = %g±%g, y0 = %g±%g, FWHM = %g±%g\r", cwl[4], swl[4], cwl[3], swl[3], cwl[5], swl[5]
	elseif (vError==0) // vError ought to be zero if we arrived here!
		MQP_Fit = TwoVoigtNoBase(cwv,x)
		cw1 = {cwv[0], cwv[1], cwv[2], cwv[3]}
		MQP_Fit1 = vpeakFWHM(cw1, x)
		cw2 = {cwv[4], cwv[5], cwv[6], cwv[7]}
		MQP_Fit2 = vpeakFWHM(cw2, x)		
		// convert total width to Gaussian width
		// this comes from multipeak fitting package
		// voigt width: An approximation with an accuracy of 0.02% 
		// from Olivero and Longbothum, 1977, JQSRT 17, 233		
		wg1 = cw1[2]/(0.5346*cw1[3]+sqrt(0.2166*cw1[3]^2+1))
		wl1 = wg1 * cw1[3]
		y01 = vPeakFWHM(cw1,cw1[1])	
		wg2 = cw2[2]/(0.5346*cw2[3]+sqrt(0.2166*cw2[3]^2+1))
		wl2 = wg2 * cw2[3]
		y02 = vPeakFWHM(cw2,cw2[1])
		
		// sigma wl and sigma wg probably have strong negative covariance.
		// error propagation (assuming the constants in the formula are uncertainty-free) would be interesting...
		printf "%s: Fitting Voigt peaks between %g and %g\r", NameOfWave(w), xmin, xmax
		printf "Peak 1: x0 = %g±%g, y0 = %g, FWHM = %g±%g, Gaussian FWHM = %g, Lorentzian FWHM = %g, area = %g±%g\r", cw1[1], swv[1], y01, cw1[2], swv[2], wg1, wl1, cw1[0], swv[0]
		printf "Peak 2: x0 = %g±%g, y0 = %g, FWHM = %g±%g, Gaussian FWHM = %g, Lorentzian FWHM = %g, area = %g±%g\r", cw2[1], swv[5], y02, cw2[2], swv[7], wg2, wl2, cw2[0], swv[4]
		// when a shape parameter is used, end up with smaller undertainty for wg, and (usually) larger propagated uncertainty for wl:
		// cwv[6]*cwv[7]*(sqrt( (swv[6]/cwv[6])^2 + (swv[7]/cwv[7])^2) - 2*mcv[6][7]/(cwv[6]*cwv[7]) )
		// set 1st coefficient to height for consistency with Gaussian and Lorentzian fits
		cw1[0] = y01
		cw2[0] = y02	
	endif
	if (doTag)
		string strTag = ""
		sprintf strTag "%s\rHeight = %g\rPos = %g\rFWHM = %g", NameOfWave(w), cw1[0], cw1[1], cw1[2]
		Tag /A=MB/B=1/C/N=MarqueePeakTag/I=0/L=2/Z=0/X=-5/Y=5 MQP_Fit1, cw1[1], strTag
		sprintf strTag "%s\rHeight = %g\rPos = %g\rFWHM = %g", NameOfWave(w), cw2[0], cw2[1], cw2[2]
		Tag /A=MB/B=1/C/N=MarqueePeakTag2/I=0/L=2/Z=0/X=5/Y=5 MQP_Fit2, cw2[1], strTag
	endif
	wOut = {cw1[0], cw1[1], cw1[2], cw2[0], cw2[1], cw2[2]}
	return wOut
end

static function DebuggerOff()
	DebuggerOptions
	int sav_debug = V_debugOnError
	DebuggerOptions debugOnError = 0
	return sav_debug
end

static function AreaXoptional(wave /Z w, wave /Z w_x, variable x1, variable x2)
	variable xmin = min(x1,x2), xmax = max(x1,x2)
	if (WaveExists(w_x))
		return areaXY(w_x, w, xmin, xmax)
	else
		return area(w, xmin, xmax)
	endif
end

static function /WAVE getEndPoints(wave w, wave /Z w_x, wave wXrange)
	Make /free/N=2 wEndPoints
	if (WaveExists(w_x))
		if (w_x[1]<w_x[0])
			Sort /R wXrange, wXrange
		endif
		FindLevel /Q /P w_x, wXrange[0]
		if (v_flag==0)
			wEndPoints[0] = V_LevelX
		endif
		FindLevel /Q /P w_x, wXrange[1]
		if (v_flag==0)
			wEndPoints[1] = V_LevelX
		endif
	else
		wEndPoints = x2pnt(w, wXrange)
	endif
	Sort wEndPoints, wEndPoints
	return wEndPoints
end

// -------------------  Normalize by area ---------------------

static function MarqueeNormaliseTraces([variable height])
	height = ParamIsDefault(height) ? 0 : height
	string tracenames, strXAxis, strTrace
	GetLastUserMenuInfo // S_value will be trace name
	tracenames = S_value
	if (cmpstr(tracenames, "All Traces")==0)
		tracenames = TraceNameList("",";",1)
	endif
	int i = 0, numTraces = ItemsInList(tracenames)
	for(i=0;i<numTraces;i++)
		strTrace = StringFromList(i,tracenames)
		strXAxis = StringByKey("XAXIS", TraceInfo("", strTrace, 0 ))
		GetMarquee /K $strXAxis
		if (height)
			normaliseTraceByMax(strTrace, x1=v_left, x2=v_right)
		else
			normaliseTraceByArea(strTrace, x1=v_left, x2=v_right)
		endif
	endfor
end

static function NormaliseAllTracesByArea()
	string tracelist = TraceNameList("",";",1+4)
	string s_trace
	int i = 0
	do
		s_trace = StringFromList(i,tracelist)
		if ( strlen(s_trace) == 0 )
			break
		endif
		normaliseTraceByArea(s_trace)
		i+=1
	while(1)
end

static function normaliseTraceByArea(string s_trace, [variable x1, variable x2])
	wave w = TraceNameToWaveRef("", s_trace)
	wave /Z w_x = XWaveRefFromTrace("", s_trace)
	if (ParamIsDefault(x1))
		NormaliseByArea(w, w_x=w_x)
	else
		NormaliseByArea(w, w_x=w_x, x1=x1, x2=x2)
	endif
	ReplaceWave trace=$s_trace, $NameOfWave(w)+"_n"
end

static function NormaliseByArea(wave w, [wave/Z w_x, variable x1, variable x2])
	variable var_area
	if (WaveExists(w_x))
		x1 = ParamIsDefault(x1) ? w_x[0] : x1
		x2 = ParamIsDefault(x2) ? w_x[numpnts(w_x)-1] : x2
		var_area = abs(areaXY(w_x, w, x1, x2))
	else
		x1 = ParamIsDefault(x1) ? leftx(w) : x1
		x2 = ParamIsDefault(x2) ? pnt2x(w,numpnts(w)-1) : x2
		var_area = abs(area(w, x1, x2))
	endif
	Duplicate /O w $NameOfWave(w)+"_n" /wave=w_norm
	w_norm = w/var_area
	printf "Duplicate /O %s %s_n\r" NameOfWave(w), NameOfWave(w)
	printf "%s_n = %s/%g\r" NameOfWave(w), NameOfWave(w), var_area
end

// -------------------  Normalize by peak ---------------------

static function MarqueeNormaliseByMax()
	string tracenames, strXAxis, strTrace
	GetLastUserMenuInfo // S_value will be trace name
	tracenames = S_value
	if (cmpstr(tracenames, "All Traces")==0)
		tracenames = TraceNameList("",";",1)
	endif
	int i = 0, numTraces = ItemsInList(tracenames)
	for(i=0;i<numTraces;i++)
		strTrace = StringFromList(i,tracenames)
		strXAxis = StringByKey("XAXIS", TraceInfo("", strTrace, 0 ))
		GetMarquee /K $strXAxis
		normaliseTraceByMax(strTrace, x1=v_left, x2=v_right)
	endfor
end

static function NormaliseAllTracesByMax()
	string tracelist = TraceNameList("",";",1+4), trace = ""
	int i = 0, numTraces = ItemsInList(tracelist)
	for(i=0;i<numTraces;i+=1)
		trace = StringFromList(i,tracelist)
		NormaliseTraceByMax(trace)
	endfor
end

static function NormaliseTraceByMax(string strTrace, [variable x1, variable x2])
	wave w = TraceNameToWaveRef("", strTrace)
	if (WaveExists(w)==0)
		return 0
	endif
	variable maxVal = NaN
	if (ParamIsDefault(x1)||ParamIsDefault(x2))
		maxVal = WaveMax(w)
	else
		maxVal = WaveMax(w, x1, x2)
	endif
	Duplicate /O w $NameOfWave(w)+"_n" /wave=w_norm
	w_norm = w/maxVal
	printf "duplicate /O %s %s_n\r" NameOfWave(w), NameOfWave(w)
	printf "%s_n = %s/%g\r" NameOfWave(w), NameOfWave(w), maxVal
	ReplaceWave trace=$strTrace, $NameOfWave(w)+"_n"
end

// ----------------------------------------------------------------

// removes first instance of w from graph 'graphStr', plots w,
// optionally vs w_x, on the same axes as the already plotted
// trace 'traceStr'
// BL_appendToSameAxes(graphStr, traceStr, w, w_x, w_rgb={r,g,b}, offset=1)
// appends w (vs w_x if w_x exists), sets color to (r,g,b) and matches y
// offset of traceStr
// Default is choose a color that contrasts with that of traceStr; matchRGB=1 forces
// color to match
// fill = 1 adds a translucent fill to zero and sends trace to back
static function appendToSameAxes(graphStr, traceStr, w, w_x, [w_rgb, offset, matchRGB, fill, lsize])
	string graphStr, traceStr
	wave /Z w, w_x, w_rgb
	int offset  // match y offset
	int matchRGB // match color of already plotted trace
	int fill // 25% fill to zero
	variable lsize
	
	offset = ParamIsDefault(offset) ? 0 : offset
	matchRGB = ParamIsDefault(matchRGB) ? 0 : matchRGB
	fill = ParamIsDefault(fill) ? 0 : fill
	lsize = ParamIsDefault(lsize) ? 1 : lsize
	
	string s_info = TraceInfo(graphStr, traceStr, 0)
	string s_Xax = StringByKey("XAXIS",s_info)
	string s_Yax = StringByKey("YAXIS",s_info)
	string s_flags = StringByKey("AXISFLAGS",s_info)
	variable flagBits = GrepString(s_flags, "/R") + 2*GrepString(s_flags, "/T")
	offset = offset ? GetNumFromModifyStr(s_info,"offset","{",1) : 0
	
	// get color of already plotted trace
	variable c0, c1, c2
	int startIndex = strsearch(s_info, ";rgb(x)=", 0)
	sscanf s_info[startIndex,strlen(s_info)-1], ";rgb(x)=(%d,%d,%d*", c0, c1, c2
	
	if (matchRGB==0 && ParamIsDefault(w_rgb)) // no color specified
		STRUCT RGBcolor RGB
		RGB.red = c0; RGB.green = c1; RGB.blue = c2
		contrastingColor(RGB)
		Make /free w_RGB = {RGB.red, RGB.green, RGB.blue}
	endif
	if (matchRGB) // this overides any specified color
		Make /free/O w_RGB = {c0,c1,c2}
	endif
		
	RemoveFromGraph /W=$graphStr/Z $NameOfWave(w)
	
	switch (flagBits)
		case 0:
			if (WaveExists(w_x))
				AppendToGraph /W=$graphStr/B=$s_Xax/L=$s_Yax/C=(w_rgb[0],w_rgb[1],w_rgb[2]) w vs w_x
			else
				AppendToGraph /W=$graphStr/B=$s_Xax/L=$s_Yax/C=(w_rgb[0],w_rgb[1],w_rgb[2]) w
			endif
			break
		case 1:
			if (WaveExists(w_x))
				AppendToGraph /W=$graphStr/B=$s_Xax/R=$s_Yax/C=(w_rgb[0],w_rgb[1],w_rgb[2]) w vs w_x
			else
				AppendToGraph /W=$graphStr/B=$s_Xax/R=$s_Yax/C=(w_rgb[0],w_rgb[1],w_rgb[2]) w
			endif
			break
		case 2:
			if (WaveExists(w_x))
				AppendToGraph /W=$graphStr/T=$s_Xax/L=$s_Yax/C=(w_rgb[0],w_rgb[1],w_rgb[2]) w vs w_x
			else
				AppendToGraph /W=$graphStr/T=$s_Xax/L=$s_Yax/C=(w_rgb[0],w_rgb[1],w_rgb[2]) w
			endif
			break
		case 3:
			if (WaveExists(w_x))
				AppendToGraph /W=$graphStr/T=$s_Xax/R=$s_Yax/C=(w_rgb[0],w_rgb[1],w_rgb[2]) w vs w_x
			else
				AppendToGraph /W=$graphStr/T=$s_Xax/R=$s_Yax/C=(w_rgb[0],w_rgb[1],w_rgb[2]) w
			endif
			break
	endswitch
	ModifyGraph /W=$graphStr offset($NameOfWave(w))={0,offset}, lsize($NameOfWave(w))=lsize
	
	if (fill)
		ModifyGraph /W=$graphStr mode($NameOfWave(w))=7, hbFill($NameOfWave(w))=2, usePlusRGB($NameOfWave(w))=1, plusRGB($NameOfWave(w))=(w_rgb[0],w_rgb[1],w_rgb[2],6554)
		ReorderTraces /W=$graphStr _back_, {$NameOfWave(w)}
	endif
end

// area, x0, FWHM, L/G
// assumes FWHM = wl * 0.5346 + sqrt( 0.2166 * wl^2 + wg^2)
// this comes from multipeak fitting package
// Voigt width: An approximation with an accuracy of 0.02% from Olivero and Longbothum, 1977, JQSRT 17, 233
function vpeakFWHM(wave w, variable x)
	Make/free w1 = {0, w[0], w[1], w[2]/(0.5346*w[3]+sqrt(0.2166*w[3]^2+1)), w[3]}
	return VoigtPeak(w1, x)
end

// area, x0, gFWHM, lFWHM
function vpeak(wave w, variable x)
	Make/free w1 = {0, w[0], w[1], w[2], w[3]/w[2]}
	return VoigtPeak(w1, x)
end

// y0, x0, FWHM
function lpeak(wave w, variable x)
	variable height = w[0], centre = w[1], FWHM = w[2]
	return height*((FWHM/2)^2)/( (x-centre)^2 + (FWHM/2)^2 )
end

// y0, x0, FWHM
function gpeak(wave w, variable x)
	variable height = w[0], centre = w[1], FWHM = w[2]
	return height*exp(-((x-centre)*2*sqrt(ln(2))/FWHM)^2)
end

// y0, x0, FWHM, y0, x0, FWHM
static function TwoGaussiansNoBase(Wave w, variable x)
	return w[0]*exp(-((x-w[1])*2*sqrt(ln(2))/w[2])^2) + w[3]*exp(-((x-w[4])*2*sqrt(ln(2))/w[5])^2)
end

// y0, x0, FWHM, y0, x0, FWHM
static function TwoLorentziansNoBase(Wave w, variable x)
	return w[0]*((w[2]/2)^2)/( (x-w[1])^2 + (w[2]/2)^2 ) + w[3]*((w[5]/2)^2)/( (x-w[4])^2 + (w[5]/2)^2 )
end

// area, centre, gFWHM, lFWHM, area, centre, gFWHM, lFWHM
static function TwoVoigtNoBase(Wave w, variable x)
	Make/free w1 = {0, w[0], w[1], w[2], w[3]/w[2]}, w2 = {0, w[4], w[5], w[6], w[7]/w[6]}
	return VoigtPeak(w1, x) + VoigtPeak(w2, x)
end

// area, centre, FWHM, L/V, area, centre, FWHM, L/V
static function TwoVoigtNoBaseFWHM(Wave w, variable x)
	Make/free w1 = {0, w[0], w[1], w[2]/(0.5346*w[3]+sqrt(0.2166*w[3]^2+1)), w[3]}, w2 = {0, w[4], w[5], w[6]/(0.5346*w[7]+sqrt(0.2166*w[7]^2+1)), w[7]}
	return VoigtPeak(w1, x) + VoigtPeak(w2, x)
end

//// area, centre, gFWHM, L/G, area, centre, gFWHM, L/G
//static function TwoVoigtNoBase(Wave w, Variable xx)
//	make/free w1={0, w[0], w[1], w[2], w[3]}, w2={0, w[4], w[5], w[6], w[7]}
//	return VoigtPeak(w1, xx) + VoigtPeak(w2, xx)
//end

static function contrastingColor(STRUCT RGBcolor &RGB)
	STRUCT HSLcolor HSL
	RGB2HSL(RGB, HSL)
	if (HSL.S<0.3)
		RGB.red = 65535;RGB.green = 0;RGB.blue = 0
		return 1
	endif
	variable hue = round(6*HSL.H)
	hue -= 6*(hue>5)
	switch (hue)
		case 0: // red
			RGB.red = 0; RGB.green = 65535; RGB.blue = 65535 // cyan
			break
		case 1: // yellow
			RGB.red = 0; RGB.green = 0; RGB.blue = 65535 // blue
			break
		case 2: // green
			RGB.red = 65535; RGB.green = 0; RGB.blue = 65535 // magenta
			break
		case 3: // cyan
			RGB.red = 65535; RGB.green = 0; RGB.blue = 0 // red
			break
		case 4: // blue
			RGB.red = 65535; RGB.green = 65535; RGB.blue = 0 // yellow
			break
		case 5: // magenta
			RGB.red = 0; RGB.green = 65535; RGB.blue = 0 // lime
			break
	endswitch
	return 1
end

// see easyrgb.com
static function RGB2HSL(STRUCT RGBcolor &rgb, STRUCT HSLcolor &hsl)
	// R, G and B input range = 0 - 65535
	// H, S and L output range = 0 - 1
	variable red = rgb.red/65535
	variable green = rgb.green/65535
	variable blue = rgb.blue/65535

	variable var_Min = min(red,green,blue)
	variable var_Max = max(red,green,blue)
	variable del_Max = var_Max - var_Min
	
	variable H,S,L
	L = (var_Max + var_Min)/ 2
	if (del_Max == 0) // grey
	    H = 0
	    S = 0
	else // Chromatic data
		S = (L < 0.5) ? del_Max/(var_Max + var_Min) : del_Max/(2 - var_Max - var_Min)
		
		variable del_R = ( (var_Max - red)/6 + del_Max/2 ) / del_Max
		variable del_G = ( (var_Max - green)/6 + del_Max/2 ) / del_Max
		variable del_B = ( (var_Max - blue)/6 + del_Max/2 ) / del_Max
		
		if (red == var_Max )
			H = del_B - del_G
		elseif (green == var_Max )
			H = (1/3) + del_R - del_B
		elseif (blue == var_Max )
			H = (2/3) + del_G - del_R
		endif
		H += (H < 0)
		H -= (H > 1)
	endif
	hsl.H = H
	hsl.S = S
	hsl.L = L
	return 1
end

static Structure HSLcolor
	float H
	float S
	float L
EndStructure